@@ -1,10 +1,5 @@
 # -*- coding: utf-8 -*-
 
-try:
-    from thread import start_new_thread, get_ident
-except ImportError:
-    # Python 3?
-    from _thread import start_new_thread, get_ident
 import threading
 import unittest
 import time
@@ -23,9 +18,6 @@ except NameError:
 
 unicode_type = type(IS_PYTHON3 and 'abc' or 'abc'.decode('ASCII'))
 
-def _wait():
-    # A crude wait/yield function not relying on synchronization primitives.
-    time.sleep(0.01)
 
 class SetupLuaRuntimeMixin(object):
 
@@ -36,6 +28,7 @@ class SetupLuaRuntimeMixin(object):
         self.lua = None
         gc.collect()
 
+
 class TestLuaRuntimeRefcounting(unittest.TestCase):
     def test_runtime_cleanup(self):
         lua = lupa.LuaRuntime()
@@ -44,6 +37,25 @@ class TestLuaRuntimeRefcounting(unittest.TestCase):
         self.assertEqual(1, lua_table[1])
         del lua_table
 
+    def test_refcycles(self):
+        def make_refcycle(runtime):
+            def use_runtime():
+                return runtime.eval('1+1')
+            runtime.globals()['use_runtime'] = use_runtime
+            runtime.globals()['some_leaky_data'] = list(range(1000))
+
+        gc.collect()
+        old_count = len(gc.get_objects())
+        for i in range(1000):
+            lua = lupa.LuaRuntime()
+            make_refcycle(lua)
+            self.assertEqual(2, lua.eval('use_runtime()'))
+            del lua, i
+        gc.collect()
+        new_count = len(gc.get_objects())
+        self.assertEqual(old_count, new_count)
+
+
 class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_eval(self):
         self.assertEqual(2, self.lua.eval('1+1'))
@@ -63,17 +75,17 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
         self.assertEqual(2, self.lua.eval('1+1'))
 
     def test_eval_error_message_decoding(self):
-        error = None
         try:
-            self.lua.eval('-üUNKNOWNöVALUEä')
+            self.lua.eval('require "UNKNOWNöMODULEäNAME"')
         except lupa.LuaError:
             error = (IS_PYTHON3 and '%s' or '%s'.decode('ASCII')) % sys.exc_info()[1]
         else:
             self.fail('expected error not raised')
-        expected_message = '[string "<python>"]:1: attempt to perform arithmetic on global \'üUNKNOWNöVALUEä\' (a nil value)'
+        expected_message = 'module \'UNKNOWNöMODULEäNAME\' not found'
         if not IS_PYTHON3:
             expected_message = expected_message.decode('UTF-8')
-        self.assertEqual(expected_message, error)
+        self.assertTrue(expected_message in error,
+                        '"%s" not found in "%s"' % (expected_message, error))
 
     def test_execute(self):
         self.assertEqual(2, self.lua.execute('return 1+1'))
@@ -188,8 +200,8 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
         self.assertEqual("test-STR!!", function(test()))
         self.assertEqual(True, called[0])
 
-    def test_eval(self):
-        function = self.lua.eval('function() return python.eval end')
+    def test_python_eval(self):
+        eval = self.lua.eval('function() return python.eval end')()
         self.assertEqual(2, eval('1+1'))
         self.assertEqual(2, self.lua.eval('python.eval("1+1")'))
 
@@ -201,6 +213,24 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
         table = self.lua.eval('{a=1, b=2, c=3}')
         self.assertEqual(0, len(table)) # as returned by Lua's "#" operator
 
+    def test_table_delattr(self):
+        table = self.lua.eval('{a=1, b=2, c=3}')
+        self.assertTrue('a' in table)
+        del table.a
+        self.assertFalse('a' in table)
+
+    def test_table_delitem(self):
+        table = self.lua.eval('{a=1, b=2, c=3}')
+        self.assertTrue('c' in table)
+        del table['c']
+        self.assertFalse('c' in table)
+
+    def test_table_delitem_special(self):
+        table = self.lua.eval('{a=1, b=2, c=3, __attr__=4}')
+        self.assertTrue('__attr__' in table)
+        del table['__attr__']
+        self.assertFalse('__attr__' in table)
+
     def test_len_table(self):
         table = self.lua.eval('{1,2,3,4, a=1, b=2, c=3}')
         self.assertEqual(4, len(table)) # as returned by Lua's "#" operator
@@ -423,6 +453,28 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
         self.assertEqual(2, table[2])
         self.assertEqual(9, len(table))
 
+    def test_setitem_array_none(self):
+        table = self.lua.eval('{1,2}')
+        get_none = self.lua.eval('function(t) return t[python.none] end')
+        self.assertEqual(2, len(table))
+        self.assertEqual(None, table[None])
+        self.assertEqual(None, get_none(table))
+        table[None] = 123
+        self.assertEqual(123, table[None])
+        self.assertEqual(123, get_none(table))
+        self.assertEqual(2, len(table))
+
+    def test_setitem_array_none_initial(self):
+        table = self.lua.eval('{1,python.none,3}')
+        get_none = self.lua.eval('function(t) return t[python.none] end')
+        self.assertEqual(3, len(table))
+        self.assertEqual(None, table[None])
+        self.assertEqual(None, get_none(table))
+        table[None] = 123
+        self.assertEqual(123, table[None])
+        self.assertEqual(123, get_none(table))
+        self.assertEqual(3, len(table))
+
     def test_setattr_table(self):
         table = self.lua.eval('{ const={ name="Pi", value=3.1415927 }, const2={ name="light speed", value=3e8 }, val=1 }')
 
@@ -509,8 +561,8 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
         self.assertNotEqual(None, stringlib.char)
 
     def test_libraries(self):
-        libraries = self.lua.eval('{require, table, io, os, math, string, debug, bit, jit}')
-        self.assertEqual(9, len(libraries))
+        libraries = self.lua.eval('{require, table, io, os, math, string, debug}')
+        self.assertEqual(7, len(libraries))
         self.assertTrue(None not in libraries)
 
     def test_callable_values(self):
@@ -556,37 +608,209 @@ class TestLuaRuntime(SetupLuaRuntimeMixin, unittest.TestCase):
         x = X()
 
         function = self.lua.eval('function(obj) return obj.a end')
-        self.assertEquals(function(x), 0)
+        self.assertEqual(function(x), 0)
         function = lua.eval('function(obj) return obj.a end')
-        self.assertEquals(function(x), 1)
+        self.assertEqual(function(x), 1)
 
         function = self.lua.eval('function(obj) return obj.__class__ end')
-        self.assertEquals(function(x), X)
+        self.assertEqual(function(x), X)
         function = lua.eval('function(obj) return obj.__class__ end')
         self.assertRaises(AttributeError, function, x)
 
         function = self.lua.eval('function(obj) return obj._a end')
-        self.assertEquals(function(x), 2)
+        self.assertEqual(function(x), 2)
         function = lua.eval('function(obj) return obj._a end')
         self.assertRaises(AttributeError, function, x)
 
         function = self.lua.eval('function(obj) return obj._X__a end')
-        self.assertEquals(function(x), 3)
+        self.assertEqual(function(x), 3)
         function = lua.eval('function(obj) return obj._X__a end')
         self.assertRaises(AttributeError, function, x)
 
         function = self.lua.eval('function(obj) return obj.a end')
-        self.assertEquals(function(x), 0)
+        self.assertEqual(function(x), 0)
         function = lua.eval('function(obj) return obj.a end')
-        self.assertEquals(function(x), 1)
+        self.assertEqual(function(x), 1)
+
+class TestAttributeHandlers(unittest.TestCase):
+    def setUp(self):
+        self.lua = lupa.LuaRuntime()
+        self.lua_handling = lupa.LuaRuntime(attribute_handlers=(self.attr_getter, self.attr_setter))
+
+        self.x, self.y = self.X(), self.Y()
+        self.d = {'a': "aval", "b": "bval", "c": "cval"}
+
+    def tearDown(self):
+        self.lua = None
+        gc.collect()
+
+    class X(object):
+        a = 0
+        a1 = 1
+        _a = 2
+        __a = 3
+
+    class Y(object):
+        a = 0
+        a1 = 1
+        _a = 2
+        __a = 3
+
+    def attr_getter(self, obj, name):
+        if not isinstance(name, unicode_type):
+            raise AttributeError('bad type for attr_name')
+        if isinstance(obj, self.X):
+            if not name.startswith('_'):
+                value = getattr(obj, name, None)
+                if value is not None:
+                    return value + 10
+                return None
+            else:
+                return "forbidden"
+        elif isinstance(obj, dict):
+            if name == "c":
+                name = "b"
+            return obj.get(name, None)
+        return None
+
+    def attr_setter(self, obj, name, value):
+        if isinstance(obj, self.Y):
+            return  # class Y is read only.
+        if isinstance(obj, self.X):
+            if name.startswith('_'):
+                return
+            if hasattr(obj, name):
+                setattr(obj, name, value)
+        elif isinstance(obj, dict):
+            if 'forbid_new' in obj and name not in obj:
+                return
+            obj[name] = value
+
+    def test_legal_arguments(self):
+        lupa.LuaRuntime(attribute_filter=None)
+        lupa.LuaRuntime(attribute_filter=len)
+        lupa.LuaRuntime(attribute_handlers=None)
+        lupa.LuaRuntime(attribute_handlers=())
+        lupa.LuaRuntime(attribute_handlers=[len, bool])
+        lupa.LuaRuntime(attribute_handlers=iter([len, bool]))
+        lupa.LuaRuntime(attribute_handlers=(None, None))
+        lupa.LuaRuntime(attribute_handlers=iter([None, None]))
+
+    def test_illegal_arguments(self):
+        self.assertRaises(
+            ValueError, lupa.LuaRuntime, attribute_filter=123)
+        self.assertRaises(
+            ValueError, lupa.LuaRuntime, attribute_handlers=(1, 2, 3, 4))
+        self.assertRaises(
+            ValueError, lupa.LuaRuntime, attribute_handlers=(1,))
+        self.assertRaises(
+            ValueError, lupa.LuaRuntime, attribute_handlers=(1, 2))
+        self.assertRaises(
+            ValueError, lupa.LuaRuntime, attribute_handlers=(1, len))
+        self.assertRaises(
+            ValueError, lupa.LuaRuntime, attribute_handlers=(len, 2))
+        self.assertRaises(
+            ValueError, lupa.LuaRuntime, attribute_handlers=(len, bool), attribute_filter=bool)
+
+    def test_attribute_setter_normal(self):
+        function = self.lua_handling.eval("function (obj) obj.a = 100 end")
+        function(self.x)
+        self.assertEqual(self.x.a, 100)
+
+    def test_attribute_setter_forbid_underscore(self):
+        function = self.lua_handling.eval("function (obj) obj._a = 100 end")
+        function(self.x)
+        self.assertEqual(self.x._a, 2)
+
+    def test_attribute_setter_readonly_object(self):
+        function = self.lua_handling.eval("function (obj) obj.a1 = 100 end")
+        function(self.y)
+        self.assertEqual(self.y.a1, 1)
+
+    def test_attribute_setter_dict_create(self):
+        function = self.lua_handling.eval("function (obj) obj['x'] = 'new' end")
+        function(self.d)
+        self.assertEquals(self.d.get('x'), 'new')
+
+    def test_attribute_setter_forbidden_dict_create(self):
+        self.d['forbid_new'] = True
+        function = self.lua_handling.eval("function (obj) obj['x'] = 'new' end")
+        function(self.d)
+        self.assertEquals(self.d.get('x'), None)
+
+    def test_attribute_setter_dict_update(self):
+        function = self.lua_handling.eval("function (obj) obj['a'] = 'new' end")
+        function(self.d)
+        self.assertEquals(self.d['a'], 'new')
+
+    def test_attribute_setter_forbidden_dict_update(self):
+        self.d['forbid_new'] = True
+        function = self.lua_handling.eval("function (obj) obj['a'] = 'new' end")
+        function(self.d)
+        self.assertEquals(self.d['a'], 'new')
+
+    def test_attribute_getter_forbid_double_underscores(self):
+        function = self.lua_handling.eval('function(obj) return obj.__name__ end')
+        self.assertEquals(function(self.x), "forbidden")
+
+        function = self.lua.eval('function(obj) return obj.__class__ end')
+        self.assertEquals(function(self.x), self.X)
+        function = self.lua_handling.eval('function(obj) return obj.__class__ end')
+        self.assertEquals(function(self.x), "forbidden")
+
+        function = self.lua.eval('function(obj) return obj._X__a end')
+        self.assertEquals(function(self.x), 3)
+        function = self.lua_handling.eval('function(obj) return obj._X__a end')
+        self.assertEquals(function(self.x), "forbidden")
+
+    def test_attribute_getter_mess_with_underscores(self):
+        function = self.lua.eval('function(obj) return obj._a end')
+        self.assertEquals(function(self.x), 2)
+        function = self.lua_handling.eval('function(obj) return obj._a end')
+        self.assertEquals(function(self.x), "forbidden")
+
+    def test_attribute_getter_replace_values(self):
+        function = self.lua.eval('function(obj) return obj.a end')
+        self.assertEquals(function(self.x), 0)
+        function = self.lua_handling.eval('function(obj) return obj.a end')
+        self.assertEquals(function(self.x), 10)
+
+        function = self.lua.eval('function(obj) return obj.a end')
+        self.assertEquals(function(self.x), 0)
+        function = self.lua_handling.eval('function(obj) return obj.a end')
+        self.assertEquals(function(self.x), 10)
+
+    def test_attribute_getter_lenient_retrieval(self):
+        function = self.lua.eval('function(obj) return obj.bad_attr end')
+        self.assertRaises(AttributeError, function, self.y)
+        function = self.lua_handling.eval('function(obj) return obj.bad_attr end')
+        self.assertEquals(function(self.y), None)
+
+    def test_attribute_getter_normal_dict_retrieval(self):
+        function = self.lua.eval('function(obj) return obj.a end')
+        self.assertEquals(function(self.d), "aval")
+        function = self.lua_handling.eval('function(obj) return obj.a end')
+        self.assertEquals(function(self.d), "aval")
+
+    def test_attribute_getter_modify_dict_retrival(self):
+        function = self.lua.eval('function(obj) return obj.c end')
+        self.assertEquals(function(self.d), "cval")
+        function = self.lua_handling.eval('function(obj) return obj.c end')
+        self.assertEquals(function(self.d), "bval")
+
+    def test_attribute_getter_lenient_dict_retrival(self):
+        function = self.lua.eval('function(obj) return obj.g end')
+        self.assertRaises(KeyError, function, self.d)
+        function = self.lua_handling.eval('function(obj) return obj.g end')
+        self.assertEquals(function(self.d), None)
 
 
 class TestPythonObjectsInLua(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_explicit_python_function(self):
-        lua_func = self.lua.eval('function(func) return table.foreach({ab="cd"}, python.as_function(func)) end')
-        def join(*args):
-            return '-'.join(args)
-        self.assertEqual('ab-cd', lua_func(join))
+        lua_func = self.lua.eval('function(func) t = {1, 5, 2, 4, 3}; table.sort(t, python.as_function(func)); return t end')
+        def compare(a, b):
+            return a < b
+        self.assertEqual([1, 2, 3, 4, 5], list(lua_func(compare)))
 
     def test_type_conversion(self):
         lua_type = self.lua.eval('type')
@@ -692,6 +916,34 @@ class TestPythonObjectsInLua(SetupLuaRuntimeMixin, unittest.TestCase):
         ''')
         self.assertEqual([None, None, None], list(values([None, None, None]).values()))
 
+    def test_python_iter_list_some_None(self):
+        values = self.lua.eval('''
+            function(L)
+                local t = {}
+                local i = 1
+                for value in python.iter(L) do
+                    t[i] = value
+                    i = i + 1
+                end
+                return t
+            end
+        ''')
+        self.assertEqual([None, 1, None], list(values([None, 1, None]).values()))
+
+    def test_python_iter_iterator(self):
+        values = self.lua.eval('''
+            function(L)
+                local t = {}
+                local i = 1
+                for value in python.iter(L) do
+                    t[i] = value
+                    i = i+1
+                end
+                return t
+            end
+        ''')
+        self.assertEqual([3, 2, 1], list(values(reversed([1, 2, 3])).values()))
+
 
 class TestLuaCoroutines(SetupLuaRuntimeMixin, unittest.TestCase):
     def test_coroutine_iter(self):
@@ -1258,10 +1510,259 @@ class TestThreading(unittest.TestCase):
         ##     image.show()
 
 
+class TestDontUnpackTuples(unittest.TestCase):
+    def setUp(self):
+        self.lua = lupa.LuaRuntime()  # default is unpack_returned_tuples=False
+
+        # Define a Python function which returns a tuple
+        # and is accessible from Lua as fun().
+        def tuple_fun():
+            return "one", "two", "three", "four"
+        self.lua.globals()['fun'] = tuple_fun
+
+    def tearDown(self):
+        self.lua = None
+        gc.collect()
+
+    def test_python_function_tuple(self):
+        self.lua.execute("a, b, c = fun()")
+        self.assertEqual(("one", "two", "three", "four"), self.lua.eval("a"))
+        self.assertEqual(None, self.lua.eval("b"))
+        self.assertEqual(None, self.lua.eval("c"))
+
+    def test_python_function_tuple_exact(self):
+        self.lua.execute("a = fun()")
+        self.assertEqual(("one", "two", "three", "four"), self.lua.eval("a"))
+
+
+class TestUnpackTuples(unittest.TestCase):
+    def setUp(self):
+        self.lua = lupa.LuaRuntime(unpack_returned_tuples=True)
+
+        # Define a Python function which returns a tuple
+        # and is accessible from Lua as fun().
+        def tuple_fun():
+            return "one", "two", "three", "four"
+        self.lua.globals()['fun'] = tuple_fun
+
+    def tearDown(self):
+        self.lua = None
+        gc.collect()
+
+    def test_python_function_tuple_expansion_exact(self):
+        self.lua.execute("a, b, c, d = fun()")
+        self.assertEqual("one", self.lua.eval("a"))
+        self.assertEqual("two", self.lua.eval("b"))
+        self.assertEqual("three", self.lua.eval("c"))
+        self.assertEqual("four", self.lua.eval("d"))
+
+    def test_python_function_tuple_expansion_extra_args(self):
+        self.lua.execute("a, b, c, d, e, f = fun()")
+        self.assertTrue(self.lua.eval("a == 'one'"))
+        self.assertTrue(self.lua.eval("b == 'two'"))
+        self.assertTrue(self.lua.eval("c == 'three'"))
+        self.assertTrue(self.lua.eval("d == 'four'"))
+        self.assertTrue(self.lua.eval("e == nil"))
+        self.assertTrue(self.lua.eval("f == nil"))
+        self.assertEqual("one", self.lua.eval("a"))
+        self.assertEqual("two", self.lua.eval("b"))
+        self.assertEqual("three", self.lua.eval("c"))
+        self.assertEqual("four", self.lua.eval("d"))
+        self.assertEqual(None, self.lua.eval("e"))
+        self.assertEqual(None, self.lua.eval("f"))
+
+    def test_python_function_tuple_expansion_missing_args(self):
+        self.lua.execute("a, b = fun()")
+        self.assertEqual("one", self.lua.eval("a"))
+        self.assertEqual("two", self.lua.eval("b"))
+
+    def test_translate_None(self):
+        """Lua does not understand None.  Should (almost) never see it."""
+        self.lua.globals()['f'] = lambda: (None, None)
+
+        self.lua.execute("x, y = f()")
+
+        self.assertEqual(None, self.lua.eval("x"))
+        self.assertEqual(self.lua.eval("x"), self.lua.eval("y"))
+        self.assertEqual(self.lua.eval("x"), self.lua.eval("z"))
+        self.assertTrue(self.lua.eval("x == y"))
+        self.assertTrue(self.lua.eval("x == z"))
+        self.assertTrue(self.lua.eval("x == nil"))
+        self.assertTrue(self.lua.eval("nil == z"))
+
+    def test_python_enumerate_list_unpacked(self):
+        values = self.lua.eval('''
+            function(L)
+                local t = {}
+                for index, a, b in python.enumerate(L) do
+                    assert(a + 30 == b)
+                    t[ index+1 ] = a + b
+                end
+                return t
+            end
+        ''')
+        self.assertEqual([50, 70, 90], list(values(zip([10, 20, 30], [40, 50, 60])).values()))
+
+    def test_python_enumerate_list_unpacked_None(self):
+        values = self.lua.eval('''
+            function(L)
+                local t = {}
+                for index, a, b in python.enumerate(L) do
+                    assert(a == nil)
+                    t[ index+1 ] = b
+                end
+                return t
+            end
+        ''')
+        self.assertEqual([3, 5], list(values(zip([None, None, None], [3, None, 5])).values()))
+
+    def test_python_enumerate_list_start(self):
+        values = self.lua.eval('''
+            function(L)
+                local t = {5,6,7}
+                for index, a, b, c in python.enumerate(L, 3) do
+                    assert(c == nil)
+                    assert(a + 10 == b)
+                    t[ index ] = a + b
+                end
+                return t
+            end
+        ''')
+        self.assertEqual([5, 6, 30, 50, 70],
+                         list(values(zip([10, 20, 30], [20, 30, 40])).values()))
+
+
+class TestMethodCall(unittest.TestCase):
+    def setUp(self):
+
+        self.lua = lupa.LuaRuntime(unpack_returned_tuples=True)
+
+        class C(object):
+            def __init__(self, x):
+                self.x = int(x)
+
+            def getx(self):
+                return self.x
+
+            def getx1(self, n):
+                return int(n) + self.x
+
+            def setx(self, v):
+                self.x = int(v)
+
+            @classmethod
+            def classmeth(cls, v):
+                return v
+
+            @staticmethod
+            def staticmeth(v):
+                return v
+
+        class D(C):
+            pass
+
+        def f():
+            return 100
+
+        def g(n):
+            return int(n) + 100
+
+        x = C(1)
+        self.lua.globals()['C'] = C
+        self.lua.globals()['D'] = D
+        self.lua.globals()['x'] = x
+        self.lua.globals()['f'] = f
+        self.lua.globals()['g'] = g
+        self.lua.globals()['d'] = { 'F': f, "G": g }
+        self.lua.globals()['bound0'] = x.getx
+        self.lua.globals()['bound1'] = x.getx1
+
+    def tearDown(self):
+        self.lua = None
+        gc.collect()
+
+    def test_method_call_as_method(self):
+        self.assertEqual(self.lua.eval("x:getx()"), 1)
+        self.assertEqual(self.lua.eval("x:getx1(2)"), 3)
+        self.lua.execute("x:setx(4)")
+        self.assertEqual(self.lua.eval("x:getx()"), 4)
+        self.assertEqual(self.lua.eval("x:getx1(2)"), 6)
+
+    def test_method_call_as_attribute(self):
+        self.assertEqual(self.lua.eval("x.getx()"), 1)
+        self.assertEqual(self.lua.eval("x.getx1(2)"), 3)
+        self.lua.execute("x.setx(4)")
+        self.assertEqual(self.lua.eval("x.getx()"), 4)
+        self.assertEqual(self.lua.eval("x.getx1(2)"), 6)
+
+    def test_method_call_mixed(self):
+        self.assertEqual(self.lua.eval("x.getx()"), 1)
+        self.assertEqual(self.lua.eval("x:getx1(2)"), 3)
+        self.assertEqual(self.lua.eval("x:getx()"), 1)
+        self.assertEqual(self.lua.eval("x.getx1(2)"), 3)
+
+        self.lua.execute("x:setx(4)")
+        self.assertEqual(self.lua.eval("x:getx()"), 4)
+        self.assertEqual(self.lua.eval("x.getx1(2)"), 6)
+        self.assertEqual(self.lua.eval("x.getx()"), 4)
+        self.assertEqual(self.lua.eval("x:getx1(2)"), 6)
+
+        self.lua.execute("x.setx(6)")
+        self.assertEqual(self.lua.eval("x.getx()"), 6)
+        self.assertEqual(self.lua.eval("x:getx()"), 6)
+        self.assertEqual(self.lua.eval("x.getx()"), 6)
+        self.assertEqual(self.lua.eval("x.getx1(2)"), 8)
+        self.assertEqual(self.lua.eval("x:getx1(2)"), 8)
+
+    def test_method_call_function_lookup(self):
+        self.assertEqual(self.lua.eval("f()"), 100)
+        self.assertEqual(self.lua.eval("g(10)"), 110)
+        self.assertEqual(self.lua.eval("d.F()"), 100)
+        self.assertEqual(self.lua.eval("d.G(9)"), 109)
+
+    def test_method_call_class_hierarchy(self):
+        self.assertEqual(self.lua.eval("C(5).getx()"), 5)
+        self.assertEqual(self.lua.eval("D(5).getx()"), 5)
+
+        self.assertEqual(self.lua.eval("C(5):getx()"), 5)
+        self.assertEqual(self.lua.eval("D(5):getx()"), 5)
+
+    def test_method_call_class_methods(self):
+        # unbound methods
+        self.assertEqual(self.lua.eval("C.getx(C(5))"), 5)
+        self.assertEqual(self.lua.eval("C.getx(D(5))"), 5)
+
+        # class/static methods
+        self.assertEqual(self.lua.eval("C:classmeth(5)"), 5)
+        self.assertEqual(self.lua.eval("C.classmeth(5)"), 5)
+        self.assertEqual(self.lua.eval("C.staticmeth(5)"), 5)
+
+    def test_method_call_bound(self):
+        self.assertEqual(self.lua.eval("bound0()"), 1)
+        self.assertEqual(self.lua.eval("bound1(3)"), 4)
+        self.assertEqual(self.lua.eval("python.eval('1 .__add__')(1)"), 2)
+        self.assertEqual(self.lua.eval("python.eval('1 .__add__')(2)"), 3)
+
+        # the following is an unfortunate side effect of the "self" removal
+        # on bound method calls:
+        self.assertRaises(TypeError, self.lua.eval, "bound1(x)")
+
 
 ################################################################################
 # tests for the FastRLock implementation
 
+try:
+    from thread import start_new_thread, get_ident
+except ImportError:
+    # Python 3?
+    from _thread import start_new_thread, get_ident
+
+
+def _wait():
+    # A crude wait/yield function not relying on synchronization primitives.
+    time.sleep(0.01)
+
+
 class TestFastRLock(unittest.TestCase):
     """Copied from CPython's test.lock_tests module
     """
@@ -1416,7 +1917,7 @@ class TestFastRLock(unittest.TestCase):
             _wait()
         self.assertEqual(len(phase), 2)
 
-    def test_different_thread_release_succeeds(self):
+    def DISABLED_test_different_thread_release_succeeds(self):
         # Lock can be released from a different thread.
         lock = self.locktype()
         lock.acquire()
@@ -1480,76 +1981,5 @@ class TestFastRLock(unittest.TestCase):
         self.assertFalse(lock._is_owned())
 
 
-class TestDontUnpackTuples(unittest.TestCase):
-    def setUp(self):
-
-        self.lua = lupa.LuaRuntime()  # default is uknpack_tuples=False
-
-        # Define a Python function which returns a tuple
-        # and is accessible from Lua as fun().
-        def tuple_fun():
-            return "one", "two", "three", "four"
-        self.lua.globals()['fun'] = tuple_fun
-
-    def tearDown(self):
-        self.lua = None
-        gc.collect()
-
-    def test_python_function_tuple(self):
-        self.lua.execute("a, b, c = fun()")
-        self.assertEqual(("one", "two", "three", "four"), self.lua.eval("a"))
-        self.assertEqual(None, self.lua.eval("b"))
-        self.assertEqual(None, self.lua.eval("c"))
-
-    def test_python_function_tuple_exact(self):
-        self.lua.execute("a = fun()")
-        self.assertEqual(("one", "two", "three", "four"), self.lua.eval("a"))
-
-
-class TestUnpackTuples(unittest.TestCase):
-    def setUp(self):
-
-        self.lua = lupa.LuaRuntime(unpack_returned_tuples=True)
-
-        # Define a Python function which returns a tuple
-        # and is accessible from Lua as fun().
-        def tuple_fun():
-            return "one", "two", "three", "four"
-        self.lua.globals()['fun'] = tuple_fun
-
-    def tearDown(self):
-        self.lua = None
-        gc.collect()
-
-    def test_python_function_tuple_expansion_exact(self):
-        self.lua.execute("a, b, c, d = fun()")
-        self.assertEqual("one", self.lua.eval("a"))
-        self.assertEqual("two", self.lua.eval("b"))
-        self.assertEqual("three", self.lua.eval("c"))
-        self.assertEqual("four", self.lua.eval("d"))
-
-    def test_python_function_tuple_expansion_extra_args(self):
-        self.lua.execute("a, b, c, d, e, f = fun()")
-        self.assertEqual("one", self.lua.eval("a"))
-        self.assertEqual("two", self.lua.eval("b"))
-        self.assertEqual("three", self.lua.eval("c"))
-        self.assertEqual("four", self.lua.eval("d"))
-        self.assertEqual(None, self.lua.eval("e"))
-        self.assertEqual(None, self.lua.eval("f"))
-
-    def test_python_function_tuple_expansion_missing_args(self):
-        self.lua.execute("a, b = fun()")
-        self.assertEqual("one", self.lua.eval("a"))
-        self.assertEqual("two", self.lua.eval("b"))
-
 if __name__ == '__main__':
-    import os
-    import unittest
-    import doctest
-    suite = unittest.TestSuite()
-    suite.addTest(doctest.DocTestSuite(lupa._lupa))
-    suite.addTest(unittest.defaultTestLoader.loadTestsFromName('__main__'))
-    suite.addTest(doctest.DocFileSuite('../../README.rst'))
-    runner = unittest.TextTestRunner(verbosity=2)
-    if not runner.run(suite).wasSuccessful():
-        sys.exit(1)
+    unittest.main()
